# -*- coding: utf-8 -*-
#
# (C) Copyright 2018 InovaDevelopment.com
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# Author: Karl  Schopmeyer <inovadevelopment.com>
#

"""
Pywbem mock tests for a complex association.  These tests are separate because
they were added later than original tests. They provide a test of the logic
of processing complex associations including ternary associations (more than
2 reference properties and subclasses of the association and result classes)
"""
from __future__ import absolute_import, print_function

import os

import pytest

from ...utils import skip_if_moftab_regenerated

# pylint: disable=wrong-import-position, wrong-import-order, invalid-name
from ...utils import import_installed
pywbem = import_installed('pywbem')
from pywbem import CIMInstanceName, CIMClassName, \
    DEFAULT_NAMESPACE  # noqa: E402
pywbem_mock = import_installed('pywbem_mock')
# pylint: enable=wrong-import-position, wrong-import-order, invalid-name

# List of initially existing namespaces in the CIM repository
INITIAL_NAMESPACES = [DEFAULT_NAMESPACE]


@pytest.fixture()
def complex_assoc_mof(tst_qualifiers_mof):
    """
    Defines a ternary association to enable testing of complex association
    model mocking. This model rougly simulates the
    CIM_InitiatorTargetLogicalUnitPath with reference properties for
    Initiator, Target, and LogicalDevice. However we kept the naming
    very simple and very limited properties in the classes to simplify the
    tests.
    """
    complex_associations_mof = """
//
//  MOF models defines a ternary association with subclasses to test
//  the capability to process a ternary association and also toe
//  process subclasses.
//  This is based on the CIM InitiatorTargetLogicalPath classes in
//  CIM with references for Initiator, Target, and LogicalDevice
//  The names have been simplified to make creating tests easier.

// This model requires the following associators which must be installed
// separately
// Association, Description, Key, Override
// The qualifiers were separated so that the MOF can be loaded into
// a running server for comparison

    [Description("Top level class i.e ManagedElement")]
class TST_ME {
        [key]
    Uint32 InstanceID;
};

    [Description("ManagedElement")]
class TST_EP:TST_ME {
    string EP_Prop;
};

    [Description("EndPoint")]
class TST_EPSub:TST_EP {
    string EP_SubProp;
};

    [Description("LogicalDevice")]
class TST_LD:TST_ME  {
    string LD_Prop;
};

    [Description("LogicalDevice Subclass")]
class TST_LDSub:TST_LD  {
    string LD_SubProp;
};

   [Association ( true ), Description ("Ternary way association." )]
class TST_A3 {
        [Key ( true ), Description ( "Initiator Endpoint." )]
    TST_EP REF Initiator;

        [Key ( true ),
         Description ( "Target endpoint." )]
    TST_EP REF Target;

        [Key ( true ),
         Description ("Subclass of LogicalDevice representing a Logical Unit" )]
    TST_LD REF LogicalUnit;};

    [Association ( true ), Description ("Ternary way association subclass." )]
class TST_A3Sub:TST_A3 {
       [Key ( true ), Description ( "Initiator Endpointsub." ),
        Override ( "Initiator" )]
    TST_EPSub REF Initiator;

       [Key ( true ), Description ( "Target endpointsub." ),
        Override ( "Target" )]
    TST_EPSub REF Target;

       [Key ( true ), Description (
           "Subclass of LogicalDevice representing a Logical Unit" ),
        Override ( "LogicalUnit" )]
    TST_LDSub REF LogicalUnit;

        [Description ( "Extra property to confirm that it does cause issues." )]
    string TST_Prop;
};

// Instances for one association
// Relate two TST_EP instances (initiator, target) to 1 TST_LD (this
// class and subclass)
// NOTE: InstanceID is monotonically increasing integer for all instances
// in the model to make displays small and simplify tests
instance of TST_EP as $EP1I {
    InstanceID = 1;
    EP_Prop = "Initiator1";};

instance of TST_EP as $EP1T {
    InstanceID = 2;
    EP_Prop = "Target1";};

instance of TST_LD as $LD1 {
    InstanceID = 3;
    LD_Prop = "LogDev1";};

// The top level association instance
instance of TST_A3 as $A311 {
    Initiator = $EP1I;
    Target = $EP1T;
    LogicalUnit = $LD1;
};

// Extra instances that are not part of basic model but that will be used
// To create more instances in test

instance of TST_LDSub as $LD4Sub {
    InstanceID = 4;
    LD_Prop = "LogDev2";};

instance of TST_EP as $EP5I {
    InstanceID = 5;
    EP_Prop = "Initiator5";};

instance of TST_EPSub as $EPSub6I {
    InstanceID = 6;
    EP_Prop = "Initiator6";
    EP_SubProp = "Subprop EPSUB6I";};

instance of TST_EPSub as $EPSub7T {
    InstanceID = 7;
    EP_Prop = "Target7";
    EP_SubProp = "Subprop EPSUB7T";};
"""
    return tst_qualifiers_mof + '\n\n' + complex_associations_mof + '\n\n'


VERBOSE = False

# Flag that when set true causes the data for tests to be executed to be
# saved in a manner that we can duplicate the test through pywbemcli and
# OpenPegasus. This includes the test mof, the request and the expected response
SAVE_TEST = False

# test variables to allow selectively executing tests.
OK = True
RUN = True
FAIL = False


PYWBEMCLI_CMDS = 'complex_assoc_pywbemcli_cmds.txt'
COMPLEX_MODEL_MOF = 'complex_association_model.mof'
COMPLEX_MODEL_OUT_MOF = 'complex_association_model_out.mof'


@pytest.fixture(scope='module', autouse=True)
def cleanup():
    """
    Cleanup the save_data files if they exist and if running tests with
    SAVE_TEST == True. This should run once per execution of the tests.
    """
    if SAVE_TEST:
        if os.path.exists(COMPLEX_MODEL_OUT_MOF):
            os.remove(COMPLEX_MODEL_OUT_MOF)
        if os.path.exists(COMPLEX_MODEL_MOF):
            os.remove(COMPLEX_MODEL_MOF)
        if os.path.exists(PYWBEMCLI_CMDS):
            os.remove(PYWBEMCLI_CMDS)


def save_data(conn, mof, request, response, exp_response):
    """
    Save the defined strings to a file as a possible pywbemcli command set
    This includes:

    1. The original mof that defined the test
    2. The mof created by the mof compiler
    3. The corresponding pywbemcli request
    4. The corresponding pywbemcli response
    5. The expected response from the test
    """
    if SAVE_TEST:
        if not os.path.exists(COMPLEX_MODEL_OUT_MOF):
            conn.display_repository(dest=COMPLEX_MODEL_OUT_MOF)

        if not os.path.exists(COMPLEX_MODEL_MOF):
            with open('complex_association_model.mof', 'w') as f:
                print(mof, file=f)

        with open(PYWBEMCLI_CMDS, 'a') as f:
            print('%s ' % request, file=f)
            print("Resp : %s" % response, file=f)
            print("ExpResp : %s" % exp_response, file=f)


X1 = """
// Instances for one association
// Relate two TST_EP instances (initiator, target) to 1 TST_LD (this
// class and subclass)
// This association replaces LogicalUnit with instance from subclass
instance of TST_A3 as $A3114 {
    Initiator = $EP1I;
    Target = $EP1T;
    LogicalUnit = $LD4Sub;
};
"""

X2 = """
// Instances for one association
// Relate two TST_EP instances (initiator, target) to 1 TST_LD (this
// class and subclass)
// NOTE: InstanceID is monotonically increasing integer for all instances
// in the model to make displays small and simplify tests
instance of TST_A3 as $A3511 {
    Initiator = $EP5I;
    Target = $EP1T;
    LogicalUnit = $LD1;
};
"""

X3 = """
// Instances for one association
// Relate two TST_EP instances (initiator, target) to 1 TST_LD (this
// class and subclass)
// NOTE: InstanceID is monotonically increasing integer for all instances
// in the model to make displays small and simplify tests
instance of TST_A3Sub as $A3Sub511 {
    Initiator = $EPSub6I;
    Target = $EPSub7T;
    LogicalUnit = $LD4Sub;
};
"""

X4 = """
// Instances for one association
// Relate two TST_EP instances (initiator, target) to 1 TST_LD (this
// class and subclass)
// NOTE: InstanceID is monotonically increasing integer for all instances
// in the model to make displays small and simplify tests
instance of TST_A3Sub as $A3Sub511 {
    Initiator = $EP1I;
    Target = $EPSub7T;
    LogicalUnit = $LD4Sub;
};
"""


@pytest.mark.parametrize(
    "ns", INITIAL_NAMESPACES + [None])
@pytest.mark.parametrize(
    "target, r, rc, mof, exp_rslt, cond", [
        # target: Target Classname (i.e. the association object name)
        # r: role attribute
        # ac: associated class attribute
        # rr: resultrole attribute
        # rc: resultclass attribute
        # mof: Extra instance MOF used for some tests
        # exp_result: Either list of names of expected classes returned
        #             or string defining error response
        # cond: True; run test; 'pdb'; start debugger; False; skip test
        # Test TST_EP  and role as initiator
        # targcln, r, rc
        # pylint: disable=line-too-long
        ['TST_EP', None, None, None, ['TST_A3'], OK],
        ['TST_EP', None, 'TST_A3', None, ['TST_A3'], OK],
        ['TST_EP', 'Initiator', 'TST_A3', None, ['TST_A3'], OK],
        ['TST_EP', 'Target', 'TST_A3', None, ['TST_A3'], OK],
        ['TST_EP', 'LogicalUnit', 'TST_A3', None, [], OK],
        ['TST_EPSub', 'Initiator', 'TST_A3', None, ['TST_A3', 'TST_A3Sub'], OK],

        ['TST_LD', None, None, None, ['TST_A3'], OK],
        ['TST_LD', None, 'TST_A3', None, ['TST_A3'], OK],
        ['TST_LD', 'LogicalUnit', 'TST_A3', None, ['TST_A3'], OK],
        ['TST_LDSub', 'LogicalUnit', 'TST_A3', None, ['TST_A3', 'TST_A3Sub'],
         OK],
    ]
)
def test_complexref_classnames(conn, ns, target, r, rc, mof, exp_rslt,
                               complex_assoc_mof, cond):
    # pylint: disable=redefined-outer-name,invalid-name
    """
    Test referencenames class operations against a ternary model defined by
    the fixture complex_assoc_mof
    """
    if not cond:
        pytest.skip("Condition for test case not met")

    skip_if_moftab_regenerated()

    mof = mof or ""
    conn.compile_mof_string(complex_assoc_mof + mof, namespace=ns)

    if ns is not None:
        target = CIMClassName(target, namespace=ns)
    if cond == 'pdb':
        # pylint: disable=import-outside-toplevel
        import pdb
        pdb.set_trace()  # pylint: disable=no-member

    rtn_clns = conn.ReferenceNames(target, ResultClass=rc, Role=r)
    exp_ns = ns or conn.default_namespace

    assert isinstance(rtn_clns, list)
    for cln in rtn_clns:
        assert isinstance(cln, CIMClassName)
        assert cln.host == conn.host
        assert cln.namespace == exp_ns

    exp_clns = [CIMClassName(classname=n, namespace=exp_ns, host=conn.host)
                for n in exp_rslt]

    request = "pywbemcli class references {0} --role {1} --result_class {2}". \
        format(target, r, rc)
    response = [c.classname for c in exp_clns]
    exp_response = [c.classname for c in exp_clns]
    save_data(conn, complex_assoc_mof, request, response, exp_response)

    if VERBOSE:
        print('\nACT %s\nEXP %s' % ([c.classname for c in rtn_clns],
                                    [c.classname for c in exp_clns]))

    assert set(cln.classname.lower() for cln in exp_clns) == \
        set(cln.classname.lower() for cln in rtn_clns)


@pytest.mark.parametrize(
    "ns", INITIAL_NAMESPACES + [None])
@pytest.mark.parametrize(
    "target, r, rc, mof, exp_rslt, cond", [
        # target: Target Classname (i.e. the association object name)
        # r: role attribute
        # ac: associated class attribute
        # rr: resultrole attribute
        # rc: resultclass attribute
        # mof: MOF for Extra instances to compile to extend test
        # exp_result: List of instance names to be returned as tuple with
        #             item 0 classname, item 1, dictionary of reference prop
        #             definitions for each reference property
        # cond: True; run test; 'pdb'; start debugger; False; skip test
        # Test TST_EP  and role as initiator
        # targcln, r, rc
        (
            ('TST_EP', 1), None, None, None,
            [
                ('TST_A3', {'Initiator': ('TST_EP', 1),
                            'Target': ('TST_EP', 2),
                            'LogicalUnit': ('TST_LD', 3)})
            ],
            OK
        ),
        (
            ('TST_EP', 1), None, 'TST_A3', None,
            [
                ('TST_A3', {'Initiator': ('TST_EP', 1),
                            'Target': ('TST_EP', 2),
                            'LogicalUnit': ('TST_LD', 3)})
            ],
            OK
        ),
        (
            ('TST_EP', 1), 'Initiator', 'TST_A3', None,
            [
                ('TST_A3', {'Initiator': ('TST_EP', 1),
                            'Target': ('TST_EP', 2),
                            'LogicalUnit': ('TST_LD', 3)})
            ],
            OK
        ),
        (
            ('TST_EP', 1), 'Target', 'TST_A3', None,
            [],
            OK
        ),
        (
            ('TST_EP', 1), 'LogicalUnit', 'TST_A3', None,
            [],
            OK
        ),
        (
            ('TST_EPSub', 6), 'Initiator', 'TST_A3', None,
            [],
            OK
        ),
        (
            ('TST_EPSub', 6), 'Initiator', 'TST_A3Sub', X3,
            [
                ('TST_A3Sub', {'Initiator': ('TST_EPSub', 6),
                               'Target': ('TST_EPSub', 7),
                               'LogicalUnit': ('TST_LDSub', 4)})
            ],
            OK
        ),
        (
            ('TST_LD', 3), None, None, None,
            [
                ('TST_A3', {'Initiator': ('TST_EP', 1),
                            'Target': ('TST_EP', 2),
                            'LogicalUnit': ('TST_LD', 3)})
            ],
            OK
        ),
        (
            ('TST_LD', 3), None, 'TST_A3', None,
            [
                ('TST_A3', {'Initiator': ('TST_EP', 1),
                            'Target': ('TST_EP', 2),
                            'LogicalUnit': ('TST_LD', 3)})
            ],
            OK
        ),
        (
            ('TST_LD', 3), 'LogicalUnit', 'TST_A3', None,
            [
                ('TST_A3', {'Initiator': ('TST_EP', 1),
                            'Target': ('TST_EP', 2),
                            'LogicalUnit': ('TST_LD', 3)})
            ],
            OK
        ),
        (
            ('TST_LDSub', 4), 'LogicalUnit', 'TST_A3Sub', X3,
            [
                ('TST_A3Sub', {'Initiator': ('TST_EPSub', 6),
                               'Target': ('TST_EPSub', 7),
                               'LogicalUnit': ('TST_LDSub', 4)})
            ],
            OK
        ),
    ]
)
def test_complexref_instnames(conn, ns, target, r, rc, mof, exp_rslt,
                              complex_assoc_mof, cond):
    # pylint: disable=redefined-outer-name,invalid-name
    """
    Test referencenames class operations against a ternary model defined by
    the fixture complex_assoc_mof
    """

    if not cond:
        pytest.skip("Condition for test case not met")

    skip_if_moftab_regenerated()

    mof = mof or ""
    conn.compile_mof_string(complex_assoc_mof + mof, namespace=ns)

    if cond == 'pdb':
        # pylint: disable=import-outside-toplevel
        import pdb
        pdb.set_trace()  # pylint: disable=no-member

    target_inst = CIMInstanceName(target[0],
                                  keybindings={'InstanceID': target[1]},
                                  namespace=ns)

    rtn_instnames = conn.ReferenceNames(target_inst, ResultClass=rc, Role=r)
    exp_ns = ns or conn.default_namespace

    assert isinstance(rtn_instnames, list)
    for instname in rtn_instnames:
        assert isinstance(instname, CIMInstanceName)
        assert instname.host == conn.host
        assert instname.namespace == exp_ns

    # Build the expected instance name to be returned. With the defined model
    # this is the association with 3 references. Each entry in exp_rslt
    # defines a classname and 3 integers representing the InstanceID of the
    # reference property
    exp_instnames = []
    if exp_rslt:
        for response_item in exp_rslt:
            kb = {}
            for prop_def, prop_value in response_item[1].items():
                kb[prop_def] = CIMInstanceName(prop_value[0],
                                               keybindings={'InstanceID':
                                                            prop_value[1]},
                                               namespace=exp_ns)
            exp_instnames.append(CIMInstanceName(response_item[0],
                                                 keybindings=kb,
                                                 namespace=exp_ns,
                                                 host=conn.host))

    rtn_instnames_str = [str(n) for n in rtn_instnames]
    exp_instnames_str = [str(n) for n in exp_instnames]
    if VERBOSE:
        print('\nACT:\n%s\nEXP:\n%s' % ("\n".join(rtn_instnames_str),
                                        "\n".join(exp_instnames_str)))

    request = "pywbemcli class references {0} --role {1} --result_class {2}". \
        format(target_inst, r, rc)
    save_data(conn, complex_assoc_mof, request, rtn_instnames_str,
              exp_instnames_str)

    assert set(exp_instnames) == set(rtn_instnames)


@pytest.mark.parametrize(
    "ns", INITIAL_NAMESPACES + [None])
@pytest.mark.parametrize(
    "target, r, ac, rr, rc, mof, exp_rslt, cond", [
        # targ: Target Classname (i.e. the association object name)
        # r: role attribute
        # ac: associated class attribute
        # rr: resultrole attribute
        # rc: resultclass attribute
        # mof: MOF for Extra instances to compile to extend test
        # exp_result: Either list of names of expected classes returned
        #             or string defining error response
        # cond: True; run test; 'pdb'; start debugger; False; skip test
        # Test TST_EP  and role as initiator
        # targcln, r, ac, rr, rc
        # pylint: disable=line-too-long
        ['TST_EP', None, None, None, None, None, ['TST_EP', 'TST_LD'], OK],
        ['TST_EP', None, 'TST_A3', None, None, None, ['TST_EP', 'TST_LD'], OK],
        ['TST_EP', 'Initiator', None, None, None, None, ['TST_EP', 'TST_LD'], OK],  # noqa: 501
        ['TST_EP', 'Initiator', 'TST_A3', None, None, None, ['TST_EP', 'TST_LD'], OK],  # noqa: 501
        ['TST_EP', None, 'TST_A3', 'Target', None, None, ['TST_EP'], OK],
        ['TST_EP', 'Initiator', 'TST_A3', 'Target', None, None, ['TST_EP'], OK],
        ['TST_EP', None, 'TST_A3', 'Target', 'TST_LD', None, [], OK],
        ['TST_EP', 'Initiator', 'TST_A3', 'Target', None, None, ['TST_EP'], OK],
        ['TST_EP', 'Initiator', 'TST_A3', 'Target', 'TST_LD', None, [], OK],
        ['TST_EP', 'Initiator', 'TST_A3', 'LogicalUnit', 'TST_LD', None, ['TST_LD'], OK],  # noqa: 501
        # test for case indepence on all parameters
        ['tst_ep', 'INITIATOR', 'tst_a3', 'LOGICALUNIT', 'tst_ld', None, ['TST_LD'], OK],  # noqa: 501

        # test source TST_EP and Target as role
        ['TST_EP', None, None, None, None, None, ['TST_EP', 'TST_LD'], OK],
        ['TST_EP', None, 'TST_A3', None, None, None, ['TST_EP', 'TST_LD'], OK],
        ['TST_EP', 'Target', None, None, None, None, ['TST_EP', 'TST_LD'], OK],
        ['TST_EP', 'Target', 'TST_A3', None, None, None, ['TST_EP', 'TST_LD'], OK],  # noqa: 501
        ['TST_EP', None, 'TST_A3', 'Initiator', None, None, ['TST_EP'], OK],
        ['TST_EP', 'Target', 'TST_A3', 'Initiator', None, None, ['TST_EP'], OK],
        ['TST_EP', None, 'TST_A3', 'Initiator', 'TST_LD', None, [], OK],
        ['TST_EP', 'Target', 'TST_A3', 'Initiator', None, None, ['TST_EP'], OK],
        ['TST_EP', 'Target', 'TST_A3', 'Initiator', 'TST_LD', None, [], OK],
        ['TST_EP', 'Target', 'TST_A3', 'LogicalUnit', 'TST_LD', None, ['TST_LD'], OK],  # noqa: 501

        # Test source TST_LD, Initiator as ResultRole
        # TODO 4 Failures. PEG Returns TST_EP only. Mock ACT TST_EP and TST_LD
        ['TST_LD', None, None, None, None, None, ['TST_EP'], OK],
        ['TST_LD', None, 'TST_A3', None, None, None, ['TST_EP'], OK],
        ['TST_LD', 'LogicalUnit', None, None, None, None, ['TST_EP'], OK],
        ['TST_LD', 'LogicalUnit', 'TST_A3', None, None, None, ['TST_EP'], OK],
        ['TST_LD', None, 'TST_A3', 'Initiator', None, None, ['TST_EP'], OK],
        # TODO: did we miss this option with all params in tests above
        ['TST_LD', 'LogicalUnit', 'TST_A3', 'Initiator', 'TST_EP', None, ['TST_EP'], OK],  # noqa: 501
        ['TST_LD', 'LogicalUnit', 'TST_A3', 'Initiator', None, None, ['TST_EP'], OK],  # noqa: 501
        ['TST_LD', None, 'TST_A3', 'Initiator', 'TST_LD', None, [], OK],
        ['TST_LD', 'LogicalUnit', 'TST_A3', 'Initiator', None, None, ['TST_EP'], OK],  # noqa: 501
        ['TST_LD', 'LogicalUnit', 'TST_A3', 'Initiator', 'TST_EP', None, ['TST_EP'], OK],  # noqa: 501

        ['TST_LD', 'LogicalUnit', 'TST_A3', 'Initiator', 'TST_LD', None, [], OK],  # noqa: 501
        ['TST_LD', 'LogicalUnit', 'TST_A3', 'Target', None, None, ['TST_EP'], OK],  # noqa: 501
        ['TST_LD', None, 'TST_A3', 'Target', 'TST_LD', None, [], OK],
        ['TST_LD', 'LogicalUnit', 'TST_A3', 'Target', None, None, ['TST_EP'], OK],  # noqa: 501
        ['TST_LD', 'LogicalUnit', 'TST_A3', 'Target', 'TST_EP', None, ['TST_EP'], OK],  # noqa: 501
        ['TST_LD', 'LogicalUnit', 'TST_A3', 'Target', 'TST_LD', None, [], OK],
        # pylint: enable=line-too-long

    ]
)
def test_complexassoc_classnames(conn, ns, target, r, rr, ac,
                                 rc, mof, exp_rslt, complex_assoc_mof, cond):
    # pylint: disable=redefined-outer-name,invalid-name
    """
    Test associatornames class operations against a ternary model defined by
    the fixture complex_assoc_mof. We do not test the associator calls
    since that logic just takes the names and expands to return instances
    """
    if not cond:
        pytest.skip("Condition for test case not met")

    skip_if_moftab_regenerated()

    mof = mof or ""
    conn.compile_mof_string(complex_assoc_mof + mof, namespace=ns)

    if ns is not None:
        target = CIMClassName(target, namespace=ns)
    if cond == 'pdb':
        # pylint: disable=import-outside-toplevel
        import pdb
        pdb.set_trace()  # pylint: disable=no-member

    rtn_clns = conn.AssociatorNames(target,
                                    AssocClass=ac,
                                    Role=r,
                                    ResultRole=rr,
                                    ResultClass=rc)
    exp_ns = ns or conn.default_namespace

    assert isinstance(rtn_clns, list)
    for cln in rtn_clns:
        assert isinstance(cln, CIMClassName)
        assert cln.host == conn.host
        assert cln.namespace == exp_ns

    exp_clns = [CIMClassName(classname=n, namespace=exp_ns, host=conn.host)
                for n in exp_rslt]
    if VERBOSE:
        print('\nACT %s\nEXP %s' % ([c.classname for c in rtn_clns],
                                    [c.classname for c in exp_clns]))

    request = "pywbemcli class associators {0} --role {1} --assoc-class {2} " \
        "--result-role {3} --result-class {4}". \
        format(target, r, ac, rr, rc)
    save_data(conn, complex_assoc_mof, request, rtn_clns, exp_clns)

    assert set(cln.classname.lower() for cln in exp_clns) == \
        set(cln.classname.lower() for cln in rtn_clns)


@pytest.mark.parametrize(
    "ns", INITIAL_NAMESPACES + [None])
@pytest.mark.parametrize(
    "target, r, ac, rr, rc, , mof, exp_rslt, cond", [
        # target: Tuple or list representing instname with 2 components
        #         (Classname, value for InstanceID key property)
        # r: role attribute
        # ac: associated class attribute
        # rr: resultrole attribute
        # rc: resultclass attribute
        # mof: MOF for Extra instances to compile to extend test
        # exp_result: Either list of names of expected classes returned
        #             or string defining error response
        # cond: True; run test; 'pdb'; start debugger; False; skip test
        # targ, r, ac, rr, rc, exp_result, cond
        # pylint: disable=line-too-long
        [
            ('TST_EP', 1), None, None, None, None, None,
            [
                ('TST_EP', 2),
                ('TST_LD', 3)
            ],
            OK
        ],
        [
            ('TST_EP', 1), None, 'TST_A3', None, None, None,
            [
                ('TST_EP', 2),
                ('TST_LD', 3)
            ],
            OK
        ],
        [
            ('TST_EP', 1), 'Initiator', None, None, None, None,
            [
                ('TST_EP', 2),
                ('TST_LD', 3)
            ],
            OK
        ],
        [
            ('TST_EP', 1), 'Initiator', 'TST_A3', None, None, None,
            [
                ('TST_EP', 2),
                ('TST_LD', 3)
            ],
            OK
        ],
        [
            ('TST_EP', 1), None, 'TST_A3', 'Target', None, None,
            [
                ('TST_EP', 2)
            ],
            OK
        ],
        [
            ('TST_EP', 1), 'Initiator', 'TST_A3', 'Target', None, None,
            [
                ('TST_EP', 2)
            ],
            OK
        ],
        [
            ('TST_EP', 1), None, 'TST_A3', 'Target', 'TST_LD', None,
            [],
            OK
        ],
        [
            ('TST_EP', 1), 'Initiator', 'TST_A3', 'Target', None, None,
            [
                ('TST_EP', 2)
            ],
            OK
        ],
        [
            ('TST_EP', 1), 'Initiator', 'TST_A3', 'Target', 'TST_LD', None,
            [],
            OK
        ],
        [
            ('TST_EP', 1), 'Initiator', 'TST_A3', 'LogicalUnit', 'TST_LD', None,
            [
                ('TST_LD', 3)
            ],
            OK
        ],

        # test source TST_EP and Target as role
        [
            ('TST_EP', 1), None, None, None, None, None,
            [
                ('TST_EP', 2),
                ('TST_LD', 3)
            ],
            OK
        ],
        [
            ('TST_EP', 1), None, 'TST_A3', None, None, None,
            [
                ('TST_EP', 2),
                ('TST_LD', 3)
            ],
            OK
        ],
        [
            ('TST_EP', 1), 'Target', None, None, None, None,
            [],
            OK
        ],
        [
            ('TST_EP', 1), 'Target', 'TST_A3', None, None, None,
            [],
            OK
        ],
        [
            ('TST_EP', 1), None, 'TST_A3', 'Initiator', None, None,
            [],
            OK
        ],
        [
            ('TST_EP', 1), 'Target', 'TST_A3', 'Initiator', None, None,
            [],
            OK
        ],
        [
            ('TST_EP', 1), None, 'TST_A3', 'Initiator', 'TST_LD', None,
            [],
            OK
        ],
        [
            ('TST_EP', 1), 'Target', 'TST_A3', 'Initiator', None, None,
            [],
            OK
        ],
        [
            ('TST_EP', 1), 'Target', 'TST_A3', 'Initiator', 'TST_LD', None,
            [],
            OK
        ],
        [
            ('TST_EP', 1), 'Target', 'TST_A3', 'LogicalUnit', 'TST_LD', None,
            [],
            OK
        ],

        # Repeat for second EP
        [
            ('TST_EP', 2), None, None, None, None, None,
            [
                ('TST_EP', 1),
                ('TST_LD', 3)
            ],
            OK
        ],
        [
            ('TST_EP', 2), None, 'TST_A3', None, None, None,
            [
                ('TST_EP', 1),
                ('TST_LD', 3)
            ],
            OK
        ],
        [
            ('TST_EP', 2), 'Initiator', None, None, None, None,
            [],
            OK
        ],
        [
            ('TST_EP', 2), 'Initiator', 'TST_A3', None, None, None,
            [],
            OK
        ],
        [
            ('TST_EP', 2), None, 'TST_A3', 'Target', None, None,
            [],
            OK
        ],
        [
            ('TST_EP', 2), 'Initiator', 'TST_A3', 'Target', None, None,
            [],
            OK
        ],
        [
            ('TST_EP', 2), None, 'TST_A3', 'Target', 'TST_LD', None,
            [],
            OK
        ],
        [
            ('TST_EP', 2), 'Initiator', 'TST_A3', 'Target', None, None,
            [],
            OK
        ],
        [
            ('TST_EP', 2), 'Initiator', 'TST_A3', 'Target', 'TST_LD', None,
            [],
            OK
        ],
        [
            ('TST_EP', 2), None, None, None, None, None,
            [
                ('TST_EP', 1),
                ('TST_LD', 3)
            ],
            OK
        ],
        [
            ('TST_EP', 2), None, 'TST_A3', None, None, None,
            [
                ('TST_EP', 1),
                ('TST_LD', 3)
            ],
            OK
        ],
        [
            ('TST_EP', 2), 'Target', None, None, None, None,
            [
                ('TST_EP', 1),
                ('TST_LD', 3)
            ],
            OK
        ],
        [
            ('TST_EP', 2), 'Target', 'TST_A3', None, None, None,
            [
                ('TST_EP', 1),
                ('TST_LD', 3)
            ],
            OK
        ],
        [
            ('TST_EP', 2), None, 'TST_A3', 'Initiator', None, None,
            [
                ('TST_EP', 1)
            ],
            OK
        ],
        [
            ('TST_EP', 2), 'Target', 'TST_A3', 'Initiator', None, None,
            [
                ('TST_EP', 1)
            ],
            OK
        ],
        [
            ('TST_EP', 2), None, 'TST_A3', 'Initiator', 'TST_LD', None,
            [],
            OK
        ],
        [
            ('TST_EP', 2), 'Target', 'TST_A3', 'Initiator', None, None,
            [
                ('TST_EP', 1)
            ],
            OK
        ],
        [
            ('TST_EP', 2), 'Target', 'TST_A3', 'Initiator', 'TST_LD', None,
            [],
            OK
        ],
        [
            ('TST_EP', 2), 'Target', 'TST_A3', 'LogicalUnit', 'TST_LD', None,
            [
                ('TST_LD', 3)
            ],
            OK
        ],

        # Test source TST_LD, Initiator as ResultRole
        [
            ('TST_LD', 3), None, None, None, None, None,
            [
                ('TST_EP', 1),
                ('TST_EP', 2)
            ],
            OK
        ],
        [
            ('TST_LD', 3), None, 'TST_A3', None, None, None,
            [
                ('TST_EP', 1),
                ('TST_EP', 2)
            ],
            OK
        ],
        [
            ('TST_LD', 3), 'LogicalUnit', None, None, None, None,
            [('TST_EP', 1), ('TST_EP', 2)],
            OK
        ],
        [
            ('TST_LD', 3), 'LogicalUnit', 'TST_A3', None, None, None,
            [
                ('TST_EP', 1),
                ('TST_EP', 2)
            ],
            OK
        ],
        [
            ('TST_LD', 3), None, 'TST_A3', 'Initiator', None, None,
            [
                ('TST_EP', 1)
            ],
            OK
        ],
        # TODO did we miss this option with all params in tests above
        [
            ('TST_LD', 3), 'LogicalUnit', 'TST_A3', 'Initiator', 'TST_EP', None,
            [
                ('TST_EP', 1)
            ],
            OK
        ],
        [
            ('TST_LD', 3), 'LogicalUnit', 'TST_A3', 'Initiator', None, None,
            [
                ('TST_EP', 1)
            ],
            OK
        ],

        [
            ('TST_LD', 3), None, 'TST_A3', 'Initiator', 'TST_EP', None,
            [
                ('TST_EP', 1)
            ],
            OK
        ],
        [
            ('TST_LD', 3), 'LogicalUnit', 'TST_A3', 'Initiator', None, None,
            [
                ('TST_EP', 1)
            ],
            OK
        ],
        [
            ('TST_LD', 3), 'LogicalUnit', 'TST_A3', 'Initiator', 'TST_EP', None,
            [
                ('TST_EP', 1)
            ],
            OK
        ],
        [
            ('TST_LD', 3), 'LogicalUnit', 'TST_A3', 'Initiator', 'TST_LD', None,
            [],
            OK
        ],
        [
            ('TST_LD', 3), 'LogicalUnit', 'TST_A3', 'Target', None, None,
            [
                ('TST_EP', 2)
            ],
            OK
        ],
        # [
        #     ('TST_LD', 3), None, 'TST_A3', 'Target', 'TST_LD', None, None,
        #     [],
        #     OK
        # ],
        [
            ('TST_LD', 3), 'LogicalUnit', 'TST_A3', 'Target', None, None,
            [
                ('TST_EP', 2)
            ],
            OK
        ],
        [
            ('TST_LD', 3), 'LogicalUnit', 'TST_A3', 'Target', 'TST_EP', None,
            [
                ('TST_EP', 2)
            ],
            OK
        ],
        [
            ('TST_LD', 3), 'LogicalUnit', 'TST_A3', 'Target', 'TST_LD', None,
            [],
            OK
        ],

        # add extra instance of association with subclass of LD
        [
            ('TST_EP', 1), None, None, None, None, X1,
            [
                ('TST_EP', 2),
                ('TST_LD', 3),
                ('TST_LDSub', 4)
            ],
            OK
        ],
        [
            ('TST_EP', 1), None, None, None, 'TST_LDSub', X1,
            [
                ('TST_LDSub', 4)
            ],
            OK
        ],

        # add extra instance of association with multiple instances of EP
        [
            ('TST_EP', 1), None, None, None, None, X2,
            [
                ('TST_EP', 2),
                ('TST_LD', 3)
            ],
            OK
        ],
        [
            ('TST_EP', 1), None, None, None, None, X2,
            [
                ('TST_EP', 2),
                ('TST_LD', 3)
            ],
            OK
        ],
        [
            ('TST_EP', 2), None, None, None, None, X2,
            [
                ('TST_EP', 1),
                ('TST_LD', 3),
                ('TST_EP', 5)
            ],
            OK
        ],
        [
            ('TST_EP', 5), None, None, None, 'TST_EP', X2,
            [
                ('TST_EP', 2)
            ],
            OK
        ],
        [
            ('TST_EP', 1), None, None, None, 'TST_EPSub', X2,
            [],
            OK
        ],

        # Test gets subclass of assoc class
        [
            ('TST_EP', 1), None, None, None, None, X4,
            [
                ('TST_EP', 2),
                ('TST_LD', 3),
                ('TST_EPSub', 7),
                ('TST_LDSub', 4)
            ],
            OK
        ],
        [
            ('TST_EP', 1), 'Initiator', None, None, None, X4,
            [
                ('TST_EP', 2),
                ('TST_LD', 3),
                ('TST_EPSub', 7),
                ('TST_LDSub', 4)
            ],
            OK
        ],

        # TODO fails becuase of TST_A3 class.  Should this be class and
        # subclasses.
        [
            ('TST_EP', 1), 'Initiator', 'TST_A3', None, None, X4,
            [
                ('TST_EP', 2),
                ('TST_LD', 3),
                ('TST_EPSub', 7),
                ('TST_LDSub', 4)
            ],
            OK
        ],

        # Test with subclass as target instance
        [
            ('TST_EPSub', 6), None, None, None, None, X3,
            [
                ('TST_EPSub', 7),
                ('TST_LDSub', 4)
            ],
            FAIL
        ],
    ]
)
def test_complexassoc_instnames(conn, ns, target, r, rr, ac,
                                rc, mof, exp_rslt, complex_assoc_mof, cond):
    # pylint: disable=redefined-outer-name,invalid-name
    """
    Test associatornames class operations against a ternary model defined by
    the fixture complex_assoc_mof
    """
    if not cond:
        pytest.skip("Condition for test case not met")

    if cond == 'pdb':
        # pylint: disable=import-outside-toplevel
        import pdb
        pdb.set_trace()  # pylint: disable=no-member

    skip_if_moftab_regenerated()

    mof = mof or ""

    conn.compile_mof_string(complex_assoc_mof + mof, namespace=ns)

    if VERBOSE:
        conn.display_repository()

    assert isinstance(target, (tuple, list))
    for rslt in exp_rslt:
        assert isinstance(rslt, (tuple, list))

    target_inst = CIMInstanceName(target[0],
                                  keybindings={'InstanceID': target[1]},
                                  namespace=ns)
    exp_ns = ns or conn.default_namespace
    exp_instnames = []
    if exp_rslt:
        for item in exp_rslt:
            assert conn.GetClass(item[0])  # Test to assure class exists
            exp_instnames.append(CIMInstanceName(item[0],
                                                 keybindings={'InstanceID':
                                                              item[1]},
                                                 namespace=exp_ns,
                                                 host=conn.host))

    rtn_instnames = conn.AssociatorNames(target_inst,
                                         AssocClass=ac,
                                         Role=r,
                                         ResultRole=rr,
                                         ResultClass=rc)

    assert isinstance(rtn_instnames, list)
    for instname in rtn_instnames:
        assert isinstance(instname, CIMInstanceName)
        assert instname.host == conn.host
        assert instname.namespace == exp_ns

    rtn_instnames_str = [str(n) for n in rtn_instnames]
    exp_instnames_str = [str(n) for n in exp_instnames]
    if VERBOSE:
        print('\nACT:\n%s\nEXP:\n%s' % ("\n".join(rtn_instnames_str),
                                        "\n".join(exp_instnames_str)))

    request = "pywbemcli class associators {0} --role {1} --assoc-class {2} " \
        "--result-role {3} --result-class {4}". \
        format(target_inst, r, ac, rr, rc)
    save_data(conn, complex_assoc_mof, request, rtn_instnames_str,
              exp_instnames_str)

    assert set(exp_instnames) == set(rtn_instnames)
    # TODO this should be in conftest.py
    # assert_equal_ciminstancenames(exp_instnames, rtn_instnames)



"""
Test the _nocasedict module.
"""

from __future__ import absolute_import

import pytest

from ..utils.pytest_extensions import simplified_test_function

# pylint: disable=wrong-import-position, wrong-import-order, invalid-name
from ...utils import import_installed
pywbem = import_installed('pywbem')
from pywbem._nocasedict import NocaseDict  # noqa: E402
# pylint: enable=wrong-import-position, wrong-import-order, invalid-name


TESTCASES_NOCASEDICT_UNNAMEDKEYS = [

    # Testcases for NocaseDict.__setitem__() and others

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * allow: Boolean controlling whether to allow unnamed keys.
    #   * key: Key to be used for the test.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    (
        "Unnamed key with unnamed keys not allowed (error)",
        dict(
            allow=False,
            key=None,
        ),
        ValueError, None, True
    ),
    (
        "Unnamed key with unnamed keys allowed",
        dict(
            allow=True,
            key=None,
        ),
        None, None, True
    ),
    (
        "String key with unnamed keys not allowed",
        dict(
            allow=False,
            key='Dog',
        ),
        None, None, True
    ),
    (
        "String key with unnamed keys allowed",
        dict(
            allow=True,
            key='Dog',
        ),
        None, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_NOCASEDICT_UNNAMEDKEYS)
@simplified_test_function
def test_NocaseDict_unnamedkeys_getitem(testcase, allow, key):
    """
    Test function for NocaseDict.__getitem__() for unnamed keys
    """

    obj = NocaseDict()

    # This is not a very practical use, but necessary to test the function
    obj.allow_unnamed_keys = True
    obj[key] = 'foo'  # Uses __setitem__()
    obj.allow_unnamed_keys = allow

    # The code to be tested
    _ = obj[key]

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_NOCASEDICT_UNNAMEDKEYS)
@simplified_test_function
def test_NocaseDict_unnamedkeys_setitem(testcase, allow, key):
    """
    Test function for NocaseDict.__setitem__() for unnamed keys
    """

    obj = NocaseDict()
    obj.allow_unnamed_keys = allow

    # The code to be tested
    obj[key] = 'foo'

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert key in obj  # Uses __contains__()


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_NOCASEDICT_UNNAMEDKEYS)
@simplified_test_function
def test_NocaseDict_unnamedkeys_delitem(testcase, allow, key):
    """
    Test function for NocaseDict.__delitem__() for unnamed keys
    """

    obj = NocaseDict()

    # This is not a very practical use, but necessary to test the function
    obj.allow_unnamed_keys = True
    obj[key] = 'foo'  # Uses __setitem__()
    obj.allow_unnamed_keys = allow

    # The code to be tested
    del obj[key]

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert key not in obj  # Uses __contains__()


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_NOCASEDICT_UNNAMEDKEYS)
@simplified_test_function
def test_NocaseDict_unnamedkeys_contains(testcase, allow, key):
    """
    Test function for NocaseDict.__contains__() for unnamed keys
    """

    obj = NocaseDict()

    # This is not a very practical use, but necessary to test the function
    obj.allow_unnamed_keys = True
    obj[key] = 'foo'  # Uses __setitem__()
    obj.allow_unnamed_keys = allow

    # The code to be tested
    assert key in obj

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_NOCASEDICT_UNNAMEDKEYS)
@simplified_test_function
def test_NocaseDict_unnamedkeys_pop(testcase, allow, key):
    """
    Test function for NocaseDict.pop() for unnamed keys
    """

    obj = NocaseDict()

    # This is not a very practical use, but necessary to test the function
    obj.allow_unnamed_keys = True
    obj[key] = 'foo'  # Uses __setitem__()
    obj.allow_unnamed_keys = allow

    # The code to be tested
    _ = obj.pop(key)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert key not in obj  # Uses __contains__()



"""
Utility functions for pywbem testing
"""

from __future__ import absolute_import, print_function

import sys
import os
import types
import six
import pytest
from ply import yacc, lex
from packaging.version import parse as parse_version


def skip_if_moftab_regenerated():
    """
    Skip the testcase if we run against an installed version of pywbem (as
    indicated by the TEST_INSTALLED environment variable), and the testcase
    would re-generate the MOF parsing table files.

    This function should be used by test cases that parse MOF files, so the
    test case is skipped in that case.

    Background:

    Pywbem uses the `ply` package to parse CIM MOF files and part of a pywbem
    installation are parsing table files named _mofparsetab.py and _moflextab.py
    that are generated by `ply`. Before pywbem 0.14.5, the version of `ply`
    module that was used to build the pywbem distribution archive needed to be
    compatible with the version of `ply` installed in the Python environment,
    otherwise `ply` attempted to re-generate these parse table files in the
    `pywbem` installation directory. Thus, if the `pywbem` installation
    directory is in the system Python, a normal user will typically encounter
    a permission denied error.

    If the installed version of `pywbem` is 0.14.5 or higher, it has tolerance
    against mismatches between these `ply` versions, by having `ply`
    re-generate the parsing tables in memory if needed, but no longer writing
    them out to the pywbem installation directory.
    """

    test_installed = os.getenv('TEST_INSTALLED')

    pywbem = import_installed('pywbem')

    try:
        # pylint: disable=import-outside-toplevel
        from pywbem import _mofparsetab, _moflextab
    except ImportError:
        if test_installed:
            # The _mofparsetab and _moflextab files should not be auto-generated
            # in this case.
            pytest.fail("Cannot run this MOF testcase against an installed "
                        "pywbem (version {0}, installed at {1}) "
                        "because that installation does not contain the "
                        "_mofparsetab.py or _moflextab.py files".
                        format(pywbem.__version__, pywbem.__file__))
        else:
            # The _mofparsetab and _moflextab files will be auto-generated.
            return

    pywbem_not_tolerant = parse_version(pywbem.__version__) <= \
        parse_version('0.14.4')  # This causes 0.14.5.devN to be tolerant

    # pylint: disable=protected-access
    mofparsetab_version = _mofparsetab._tabversion
    moflextab_version = _moflextab._tabversion
    # pylint: enable=protected-access

    mofparsetab_mismatch = parse_version(mofparsetab_version) != \
        parse_version(yacc.__tabversion__)
    moflextab_mismatch = parse_version(moflextab_version) != \
        parse_version(lex.__tabversion__)

    if test_installed and pywbem_not_tolerant and \
            (mofparsetab_mismatch or moflextab_mismatch):
        pytest.skip("Cannot run this MOF testcase against an installed "
                    "pywbem (version {0}, installed at {1}) because that "
                    "pywbem version does not tolerate table version mismatches "
                    "between the current ply package and the generated pywbem "
                    "mof*tab files: yacc table version: current ply: {2}, "
                    "_mofparsetab.py: {3}, lex table version: current ply: "
                    "{4}, _moflextab.py: {5}".
                    format(pywbem.__version__, pywbem.__file__,
                           yacc.__tabversion__, mofparsetab_version,
                           lex.__tabversion__, moflextab_version))


def import_installed(module_name):
    """
    Import a Python module/package, controlling whether it is loaded from the
    normal Python module search path, or from an installed version (excluding
    the module in the current directory).

    The TEST_INSTALLED environment variable controls this as follows:

      * If not set or empty, the normal Python module search path is used.
        Because that search path contains the current directory in front of the
        list, this will cause a module directory in the current directory to
        have precedence over any installed versions of the module.

      * If non-empty, the current directory is removed from the Python module
        search path, and an installed version of the module is thus used, even
        when a module directory exists in the current directory. This can be
        used for testing an OS-installed version of the module.

    Example usage, e.g. in a pywbem test program::

        from ...utils import import_installed
        pywbem = import_installed('pywbem')  # pylint: disable=invalid-name
        from pywbem import ...

    The number of dots in `from ..utils` depends on where the test program
    containing this code is located, relative to the tests/utils.py file.
    """
    test_installed = os.getenv('TEST_INSTALLED')
    if test_installed:

        # Remove '' directory.
        dirpath = ''
        try:
            ix = sys.path.index(dirpath)
        except ValueError:
            ix = None
        if ix is not None:
            if test_installed == 'DEBUG':
                print("Debug: Removing {0} at index {1} from module search "
                      "path".format(dirpath, ix))
            del sys.path[ix]

        # Move CWD to end. Reason is that when testing with an editable
        # installation, the CWD is needed, but when testing with a non-editable
        # installation, the package should not be found inthe CWD.
        # Note that somehow the CWD gets inserted at the begin of the search
        # path every time, so we need a loop.
        dirpath = os.getcwd()
        while True:
            try:
                ix = sys.path.index(dirpath)
            except ValueError:
                if test_installed == 'DEBUG':
                    print("Debug: Appending {0} to end of module search "
                          "path".format(dirpath))
                sys.path.append(dirpath)
                break
            if ix == len(sys.path) - 1:
                # it exists once at the end
                break
            if test_installed == 'DEBUG':
                print("Debug: Removing {0} at index {1} from module search "
                      "path".format(dirpath, ix))
            del sys.path[ix]

    if module_name not in sys.modules:
        module = __import__(module_name, level=0)  # only absolute imports
        if test_installed == 'DEBUG':
            print("Debug: {0} module newly loaded from: {1}".
                  format(module_name, module.__file__))
    else:
        module = sys.modules[module_name]
        if test_installed == 'DEBUG':
            print("Debug: {0} module was already loaded from: {1}".
                  format(module_name, module.__file__))
    return module


def is_inherited_from(member_name, derived_class, base_class):
    """
    Return whether the specified member of a derived class was inherited from
    a base class without being overwritten in between.

    This can be used in situations where a derived class is used to test the
    (inherited) members of an abstract base class, to assert that the member
    that is used in the test is actually the inherited base class member.

    The member can be any of:

    * an instance method,
    * a static method (i.e. a method decorated with @staticmethod),
    * a property (i.e. a getter method decorated with @property).

    Data attributes are not supported at this point.

    Note that the base class may have inherited the member from a further base
    class; but that is irrelevant for this check.

    If the derived class or base class does not expose a member with the
    specified name, AttributeError is raised.
    If the type of member is not supported, TypeError is raised.

    Parameters:

      member_name (string): The name of the class member to be checked.

      derived_class (class): The derived class to be used for the check.

      base_class (class): The base class to be used for the check.

    Returns:

      bool: Boolean indicating whether the derived class member was inherited
        from the base class.
    """

    assert isinstance(member_name, six.string_types)
    assert isinstance(base_class, type)
    assert isinstance(derived_class, type)

    derived_member = getattr(derived_class, member_name)
    base_member = getattr(base_class, member_name)

    if isinstance(derived_member, types.MethodType) and \
            isinstance(base_member, types.MethodType):
        # instance method
        derived_code = derived_member.__func__.__code__
        base_code = base_member.__func__.__code__
    elif isinstance(derived_member, types.FunctionType) and \
            isinstance(base_member, types.FunctionType):
        # static method
        derived_code = derived_member.__code__
        base_code = base_member.__code__
    elif isinstance(derived_member, property) and \
            isinstance(base_member, property):
        # property
        derived_code = derived_member.fget.__code__
        base_code = base_member.fget.__code__
    else:
        raise TypeError(
            "The type of class member {} is not an instance method, static "
            "method, or a property, but is {} in derived class {} and {} in "
            "base class {}".
            format(member_name, type(derived_member), derived_class.__name__,
                   type(base_member), base_class.__name__))

    return derived_code.co_filename == base_code.co_filename and \
        derived_code.co_firstlineno == base_code.co_firstlineno



"""
Test functions in _units module
"""

from __future__ import absolute_import

import pytest

from ..utils.pytest_extensions import simplified_test_function

# pylint: disable=wrong-import-position, wrong-import-order, invalid-name
from ...utils import import_installed
pywbem = import_installed('pywbem')
from pywbem import CIMProperty, CIMMethod, CIMParameter, CIMQualifier, \
    siunit_obj, siunit  # noqa: E402
# pylint: enable=wrong-import-position, wrong-import-order, invalid-name


# Some qualifier objects used for test cases
TEST_QUAL_PUNIT1_BYTE = CIMQualifier('PUnit', value='byte')
TEST_QUAL_PUNIT2_BYTE = CIMQualifier('pUnIt', value='byte')
TEST_QUAL_UNITS1_BYTES = CIMQualifier('Units', value='Bytes')
TEST_QUAL_UNITS2_BYTES = CIMQualifier('uNiTs', value='Bytes')
TEST_QUAL_UNITS_METERS = CIMQualifier('Units', value='Meters')


TESTCASES_SIUNIT_OBJ = [

    # Testcases for siunit_obj(), that test general behavior.

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * in_kwargs: Keyword arguments for siunit().
    #   * exp_result: Expected return value of siunit(), or None.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    (
        "No input argument (cim_obj is required)",
        dict(
            in_kwargs=dict(),
            exp_result=None,
        ),
        TypeError, None, True
    ),
    (
        "cim_obj is None",
        dict(
            in_kwargs=dict(
                cim_obj=None,
            ),
            exp_result=None,
        ),
        TypeError, None, True
    ),
    (
        "cim_obj has invalid type",
        dict(
            in_kwargs=dict(
                cim_obj=42,
            ),
            exp_result=None,
        ),
        TypeError, None, True
    ),
    (
        "cim_obj is CIMProperty with PUnit #1 (byte)",
        dict(
            in_kwargs=dict(
                cim_obj=CIMProperty(
                    'P1', value='foo',
                    qualifiers=[TEST_QUAL_PUNIT1_BYTE])
            ),
            exp_result=u'B',
        ),
        None, None, True
    ),
    (
        "cim_obj is CIMMethod with PUnit #1 (byte)",
        dict(
            in_kwargs=dict(
                cim_obj=CIMMethod(
                    'P1', return_type='uint32',
                    qualifiers=[TEST_QUAL_PUNIT1_BYTE])
            ),
            exp_result=u'B',
        ),
        None, None, True
    ),
    (
        "cim_obj is CIMParameter with PUnit #1 (byte)",
        dict(
            in_kwargs=dict(
                cim_obj=CIMParameter(
                    'P1', type='uint32',
                    qualifiers=[TEST_QUAL_PUNIT1_BYTE])
            ),
            exp_result=u'B',
        ),
        None, None, True
    ),
    (
        "cim_obj is CIMProperty with PUnit #2 (byte)",
        dict(
            in_kwargs=dict(
                cim_obj=CIMProperty(
                    'P1', value='foo',
                    qualifiers=[TEST_QUAL_PUNIT2_BYTE])
            ),
            exp_result=u'B',
        ),
        None, None, True
    ),
    (
        "cim_obj is CIMProperty with Units #1 (Bytes)",
        dict(
            in_kwargs=dict(
                cim_obj=CIMProperty(
                    'P1', value='foo',
                    qualifiers=[TEST_QUAL_UNITS1_BYTES])
            ),
            exp_result=u'B',
        ),
        None, None, True
    ),
    (
        "cim_obj is CIMProperty with Units #2 (Bytes)",
        dict(
            in_kwargs=dict(
                cim_obj=CIMProperty(
                    'P1', value='foo',
                    qualifiers=[TEST_QUAL_UNITS2_BYTES])
            ),
            exp_result=u'B',
        ),
        None, None, True
    ),
    (
        "cim_obj is CIMProperty with PUnit #1 (byte) and Units (Meters)",
        dict(
            in_kwargs=dict(
                cim_obj=CIMProperty(
                    'P1', value='foo',
                    qualifiers=[TEST_QUAL_PUNIT1_BYTE, TEST_QUAL_UNITS_METERS])
            ),
            exp_result=u'B',
        ),
        None, None, True
    ),
    (
        "cim_obj is CIMProperty with Units (Meters) and PUnit #1 (byte)",
        dict(
            in_kwargs=dict(
                cim_obj=CIMProperty(
                    'P1', value='foo',
                    qualifiers=[TEST_QUAL_UNITS_METERS, TEST_QUAL_PUNIT1_BYTE])
            ),
            exp_result=u'B',
        ),
        None, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_SIUNIT_OBJ)
@simplified_test_function
def test_siunit_obj(testcase, in_kwargs, exp_result):
    """
    Test function for siunit_obj().
    """

    # The code to be tested
    result = siunit_obj(**in_kwargs)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert result == exp_result


TESTCASES_SIUNIT_GENERAL = [

    # Testcases for siunit(), that test general behavior.

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * in_kwargs: Keyword arguments for siunit().
    #   * exp_result: Expected return value of siunit(), or None.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    (
        "No input arguments",
        dict(
            in_kwargs=dict(),
            exp_result=None,
        ),
        None, None, True
    ),
    (
        "PUnit '' (empty string)",
        dict(
            in_kwargs=dict(
                punit="",
            ),
            exp_result=u'',
        ),
        None, None, True
    ),
    (
        "PUnit 'byte' as Unicode string",
        dict(
            in_kwargs=dict(
                punit=u"byte",
            ),
            exp_result=u'B',
        ),
        None, None, True
    ),
    (
        "PUnit 'byte' as binary string",
        dict(
            in_kwargs=dict(
                punit=b"byte",
            ),
            exp_result=u'B',
        ),
        None, None, True
    ),
    (
        "Units 'Bytes' as Unicode string",
        dict(
            in_kwargs=dict(
                units=u"Bytes",
            ),
            exp_result=u'B',
        ),
        None, None, True
    ),
    (
        "Units 'Bytes' as binary string",
        dict(
            in_kwargs=dict(
                units=b"Bytes",
            ),
            exp_result=u'B',
        ),
        None, None, True
    ),
    (
        "PUnit has precedence over Units",
        dict(
            in_kwargs=dict(
                punit="byte",
                units="Meters",
            ),
            exp_result=u'B',
        ),
        None, None, True
    ),
    (
        "PUnit 'degree' with use_ascii omitted is returned as Unicode char",
        dict(
            in_kwargs=dict(
                punit="degree",
            ),
            exp_result=u'\u00B0',
        ),
        None, None, True
    ),
    (
        "PUnit 'degree' with use_ascii=False is returned as Unicode char",
        dict(
            in_kwargs=dict(
                punit="degree",
                use_ascii=False,
            ),
            exp_result=u'\u00B0',
        ),
        None, None, True
    ),
    (
        "PUnit 'degree' with use_ascii=True is returned as 7-bit ASCII text",
        dict(
            in_kwargs=dict(
                punit="degree",
                use_ascii=True,
            ),
            exp_result=u'deg',
        ),
        None, None, True
    ),
    (
        "PUnit 'degree celsius' with use_ascii=True",
        dict(
            in_kwargs=dict(
                punit="degree celsius",
                use_ascii=True,
            ),
            exp_result=u'degC',
        ),
        None, None, True
    ),
    (
        "PUnit 'degree fahrenheit' with use_ascii=True",
        dict(
            in_kwargs=dict(
                punit="degree fahrenheit",
                use_ascii=True,
            ),
            exp_result=u'degF',
        ),
        None, None, True
    ),
    (
        "PUnit 'ohm' with use_ascii=True",
        dict(
            in_kwargs=dict(
                punit="ohm",
                use_ascii=True,
            ),
            exp_result=u'Ohm',
        ),
        None, None, True
    ),
    (
        "PUnit 'permille' with use_ascii=True",
        dict(
            in_kwargs=dict(
                punit="permille",
                use_ascii=True,
            ),
            exp_result=u'1/1000',
        ),
        None, None, True
    ),
    (
        "PUnit 'meter * meter' with use_ascii=True",
        dict(
            in_kwargs=dict(
                punit="meter * meter",
                use_ascii=True,
            ),
            exp_result=u'm^2',
        ),
        None, None, True
    ),
    (
        "PUnit 'meter * meter * meter' with use_ascii=True",
        dict(
            in_kwargs=dict(
                punit="meter * meter * meter",
                use_ascii=True,
            ),
            exp_result=u'm^3',
        ),
        None, None, True
    ),
    (
        "Invalid type for punit argument",
        dict(
            in_kwargs=dict(
                punit=42,
            ),
            exp_result=None,
        ),
        TypeError, None, True
    ),
    (
        "Invalid type for units argument",
        dict(
            in_kwargs=dict(
                units=42,
            ),
            exp_result=None,
        ),
        TypeError, None, True
    ),
    (
        "Invalid PUnit format",
        dict(
            in_kwargs=dict(
                punit="byte + byte",
            ),
            exp_result=None,
        ),
        ValueError, None, True
    ),
    (
        "Invalid PUnit base unit",
        dict(
            in_kwargs=dict(
                punit="peanuts",
            ),
            exp_result=None,
        ),
        ValueError, None, True
    ),
    (
        "PUnit with modifier1 (* unknown value)",
        dict(
            in_kwargs=dict(
                punit="second * 10000",
            ),
            exp_result=u'10000 s',
        ),
        None, None, True
    ),
    (
        "PUnit with modifier1 (/ unknown value)",
        dict(
            in_kwargs=dict(
                punit="second / 10000",
            ),
            exp_result=u'1/10000 s',
        ),
        None, None, True
    ),
    (
        "PUnit with modifier2 (* unknown value)",
        dict(
            in_kwargs=dict(
                punit="second * 10 ^ 4",
            ),
            exp_result=u'10^4 s',
        ),
        None, None, True
    ),
    (
        "PUnit with modifier2 (/ unknown value)",
        dict(
            in_kwargs=dict(
                punit="second / 10 ^ 4",
            ),
            exp_result=u'10^-4 s',
        ),
        None, None, True
    ),
    (
        "PUnit with modifier1 (* known value) and modifier2 (* known + exp)",
        dict(
            in_kwargs=dict(
                punit="second * 10 * 10 ^ 3",
            ),
            exp_result=u'10 ks',
        ),
        None, None, True
    ),
    (
        "PUnit with modifier1 (/ known value) and modifier2 (/ known + exp)",
        dict(
            in_kwargs=dict(
                punit="second / 10 / 10 ^ 3",
            ),
            exp_result=u'1/10 ms',
        ),
        None, None, True
    ),
    (
        "PUnit with modifier1 (* known value) and modifier2 (* known - exp)",
        dict(
            in_kwargs=dict(
                punit="second * 10 * 10 ^ -3",
            ),
            exp_result=u'10 ms',
        ),
        None, None, True
    ),
    (
        "PUnit with modifier1 (/ known value) and modifier2 (/ known - exp)",
        dict(
            in_kwargs=dict(
                punit="second / 10 / 10 ^ -3",
            ),
            exp_result=u'1/10 ks',
        ),
        None, None, True
    ),
    (
        "PUnit with modifier1 (* known value) and modifier2 (* unknown value)",
        dict(
            in_kwargs=dict(
                punit="second * 1000 * 10 ^ 4",
            ),
            exp_result=u'1000 10^4 s',
        ),
        None, None, True
    ),
    (
        "PUnit with modifier1 (/ known value) and modifier2 (/ unknown value)",
        dict(
            in_kwargs=dict(
                punit="second / 1000 / 10 ^ 4",
            ),
            exp_result=u'1/1000 10^-4 s',
        ),
        None, None, True
    ),
    (
        "PUnit with modifier1 (* unknown value) and modifier2 (* known value)",
        dict(
            in_kwargs=dict(
                punit="second * 10000 * 10 ^ 3",
            ),
            exp_result=u'10000 ks',
        ),
        None, None, True
    ),
    (
        "PUnit with modifier1 (/ unknown value) and modifier2 (/ known value)",
        dict(
            in_kwargs=dict(
                punit="second / 10000 / 10 ^ 3",
            ),
            exp_result=u'1/10000 ms',
        ),
        None, None, True
    ),
    (
        "Invalid Units value",
        dict(
            in_kwargs=dict(
                units="Peanuts",
            ),
            exp_result=None,
        ),
        ValueError, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_SIUNIT_GENERAL)
@simplified_test_function
def test_siunit_general(testcase, in_kwargs, exp_result):
    """
    Test function for siunit(), that test general behavior.
    """

    # The code to be tested
    result = siunit(**in_kwargs)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert result == exp_result


TESTCASES_SIUNIT_PUNIT = [

    # Testcases for siunit(), that test only good cases with PUnit values.

    # Each list item is a testcase tuple with these items:
    # * punit: Input value for punit parameter of siunit().
    # * exp_result: Expected return value of siunit().

    # All single units
    ("percent", u"%"),
    ("permille", u"\u2030"),
    ("decibel", u"dB"),
    ("count", u""),
    ("revolution", u"rev"),
    ("degree", u"\u00B0"),
    ("radian", u"rad"),
    ("steradian", u"sr"),
    ("bit", u"bit"),
    ("byte", u"B"),
    ("dataword", u"word"),
    ("MSU", u"MSU"),
    ("meter", u"m"),
    ("inch", u"in"),
    ("rack unit", u"U"),
    ("foot", u"ft"),
    ("yard", u"yd"),
    ("mile", u"mi"),
    ("liter", u"l"),
    ("fluid ounce", u"fl.oz"),
    ("liquid gallon", u"gal"),
    ("mole", u"mol"),
    ("kilogram", u"kg"),
    ("ounce", u"oz"),
    ("pound", u"lb"),
    ("second", u"s"),
    ("minute", u"min"),
    ("hour", u"h"),
    ("day", u"d"),
    ("week", u"week"),
    ("gravity", u"g"),
    ("degree celsius", u"\u00B0C"),
    ("degree fahrenheit", u"\u00B0F"),
    ("kelvin", u"K"),
    ("candela", u"cd"),
    ("lumen", u"lm"),
    ("nit", u"nit"),
    ("lux", u"lx"),
    ("newton", u"N"),
    ("pascal", u"Pa"),
    ("bar", u"bar"),
    ("decibel(A)", u"dB(A)"),
    ("decibel (A)", u"dB(A)"),
    ("decibel(C)", u"dB(C)"),
    ("decibel (C)", u"dB(C)"),
    ("joule", u"J"),
    ("watt", u"W"),
    ("volt ampere", u"VA"),
    ("volt ampere reactive", u"var"),
    ("decibel(m)", u"dBm"),
    ("decibel (m)", u"dBm"),
    ("british thermal unit", u"BTU"),
    ("ampere", u"A"),
    ("coulomb", u"C"),
    ("volt", u"V"),
    ("farad", u"F"),
    ("ohm", u"\u2126"),
    ("siemens", u"S"),
    ("weber", u"Wb"),
    ("tesla", u"T"),
    ("henry", u"H"),
    ("becquerel", u"Bq"),
    ("gray", u"Gy"),
    ("sievert", u"Sv"),

    # Multiple same units combined (using powers), without modifiers
    ("meter*meter", u"m\u00B2"),
    ("meter * meter", u"m\u00B2"),
    ("meter * meter * meter", u"m\u00B3"),
    ("meter * meter * meter * meter", u"m^4"),
    ("meter / second / second", u"m/s\u00B2"),
    ("meter * meter / second / second / second", u"m\u00B2/s\u00B3"),
    ("meter * meter * meter / second / second / second / second",
     u"m\u00B3/s^4"),

    # Multiple units combined, without powers, without modifiers
    ("meter/second", u"m/s"),
    ("meter / second", u"m/s"),
    ("newton * meter", u"Nm"),
    ("newton * meter / second", u"Nm/s"),
    ("newton * meter / second / byte", u"Nm/sB"),
    ("meter * newton / byte / second", u"mN/Bs"),

    # Using modifier1
    ("byte*1024", u"KiB"),
    ("byte * 1024", u"KiB"),
    ("meter / second * 1000", u"km/s"),
    ("meter / second * 100", u"hm/s"),
    ("meter / second * 10", u"dam/s"),
    ("meter / second * 1", u"m/s"),
    ("meter / second / 1", u"m/s"),
    ("meter / second / 10", u"dm/s"),
    ("meter / second / 100", u"cm/s"),
    ("meter / second / 1000", u"mm/s"),

    # Using modifier2
    ("byte*10^3", u"kB"),
    ("byte * 10^3", u"kB"),
    ("byte * 10 ^ 3", u"kB"),
    ("byte * 2 ^ 80", u"YiB"),
    ("byte * 2 ^ 70", u"ZiB"),
    ("byte * 2 ^ 60", u"EiB"),
    ("byte * 2 ^ 50", u"PiB"),
    ("byte * 2 ^ 40", u"TiB"),
    ("byte * 2 ^ 30", u"GiB"),
    ("byte * 2 ^ 20", u"MiB"),
    ("byte * 2 ^ 10", u"KiB"),
    ("byte * 10 ^ 24", u"YB"),
    ("byte * 10 ^ 21", u"ZB"),
    ("byte * 10 ^ 18", u"EB"),
    ("byte * 10 ^ 15", u"PB"),
    ("byte * 10 ^ 12", u"TB"),
    ("byte * 10 ^ 9", u"GB"),
    ("byte * 10 ^ 6", u"MB"),
    ("byte * 10 ^ 3", u"kB"),
    ("byte * 10 ^ 2", u"hB"),
    ("byte * 10 ^ 1", u"daB"),
    ("byte * 10 ^ -1", u"dB"),
    ("byte * 10 ^ -2", u"cB"),
    ("byte * 10 ^ -3", u"mB"),
    ("byte * 10 ^ -6", u"uB"),
    ("byte * 10 ^ -9", u"nB"),
    ("byte * 10 ^ -12", u"pB"),
    ("byte * 10 ^ -15", u"fB"),
    ("byte * 10 ^ -18", u"aB"),
    ("byte * 10 ^ -21", u"zB"),
    ("byte * 10 ^ -24", u"yB"),

    # Using modifier1 and modifier2
    ("byte*2*10^3", u"2 kB"),
    ("byte * 2 * 10 ^ 3", u"2 kB"),
    ("byte * 10 * 10 ^ 3", u"10 kB"),
]


@pytest.mark.parametrize(
    "punit, exp_result",
    TESTCASES_SIUNIT_PUNIT)
def test_siunit_punit(punit, exp_result):
    """
    Test function for siunit(), that test only good cases with PUnit values.
    """

    # The code to be tested
    result = siunit(punit=punit)

    assert result == exp_result


TESTCASES_SIUNIT_UNITS = [

    # Testcases for siunit(), that test only good cases with Units values.

    # Each list item is a testcase tuple with these items:
    # * units: Input value for units parameter of siunit().
    # * exp_result: Expected return value of siunit().

    # All units defined in DSP0004

    ("Bits", u"bit"),
    ("KiloBits", u"kbit"),
    ("MegaBits", u"Mbit"),
    ("GigaBits", u"Gbit"),

    ("Bits per Second", u"bit/s"),
    ("KiloBits per Second", u"kbit/s"),
    ("MegaBits per Second", u"Mbit/s"),
    ("GigaBits per Second", u"Gbit/s"),

    ("Bytes", u"B"),
    ("KiloBytes", u"kB"),
    ("MegaBytes", u"MB"),
    ("GigaBytes", u"GB"),
    ("Words", u"words"),
    ("DoubleWords", u"doublewords"),
    ("QuadWords", u"quadwords"),

    ("Degrees C", u"\u00B0C"),
    ("Tenths of Degrees C", u"1/10 \u00B0C"),
    ("Hundredths of Degrees C", u"1/100 \u00B0C"),
    ("Degrees F", u"\u00B0F"),
    ("Tenths of Degrees F", u"1/10 \u00B0F"),
    ("Hundredths of Degrees F", u"1/100 \u00B0F"),
    ("Degrees K", u"K"),
    ("Tenths of Degrees K", u"1/10 K"),
    ("Hundredths of Degrees K", u"1/100 K"),
    ("Color Temperature", u"K"),

    ("Volts", u"V"),
    ("MilliVolts", u"mV"),
    ("Tenths of MilliVolts", u"1/10 mV"),
    ("Amps", u"A"),
    ("MilliAmps", u"mA"),
    ("Tenths of MilliAmps", u"1/10 mA"),
    ("Watts", u"W"),
    ("MilliWattHours", u"mWh"),

    ("Joules", u"J"),
    ("Coulombs", u"C"),
    ("Newtons", u"N"),

    ("Lumen", u"lm"),
    ("Lux", u"lx"),
    ("Candelas", u"cd"),

    ("Pounds", u"lb"),
    ("Pounds per Square Inch", u"lb/in\u00B2"),

    ("Cycles", u"cycles"),
    ("Revolutions", u"rev"),
    ("Revolutions per Minute", u"RPM"),
    ("Revolutions per Second", u"rev/s"),

    ("Minutes", u"min"),
    ("Seconds", u"s"),
    ("Tenths of Seconds", u"1/10 s"),
    ("Hundredths of Seconds", u"1/100 s"),
    ("MicroSeconds", u"\u00B5s"),
    ("MilliSeconds", u"ms"),
    ("NanoSeconds", u"ns"),

    ("Hours", u"h"),
    ("Days", u"d"),
    ("Weeks", u"week"),

    ("Hertz", u"Hz"),
    ("MegaHertz", u"MHz"),

    ("Pixels", u"px"),
    ("Pixels per Inch", u"px/in"),

    ("Counts per Inch", u"counts/in"),

    ("Percent", u"%"),
    ("Tenths of Percent", u"1/10 %"),
    ("Hundredths of Percent", u"1/100 %"),
    ("Thousandths", u"\u2030"),

    ("Meters", u"m"),
    ("Centimeters", u"cm"),
    ("Millimeters", u"mm"),
    ("Cubic Meters", u"m\u00B3"),
    ("Cubic Centimeters", u"cm\u00B3"),
    ("Cubic Millimeters", u"mm\u00B3"),

    ("Inches", u"in"),
    ("Feet", u"ft"),
    ("Cubic Inches", u"in\u00B3"),
    ("Cubic Feet", u"ft\u00B3"),
    ("Ounces", u"oz"),
    ("Liters", u"l"),
    ("Fluid Ounces", u"fl.oz"),

    ("Radians", u"rad"),
    ("Steradians", u"sr"),
    ("Degrees", u"\u00B0"),

    ("Gravities", u"g"),
    ("Pounds", u"lb"),
    ("Foot-Pounds", u"ft.lb"),

    ("Gauss", u"G"),
    ("Gilberts", u"Gb"),
    ("Henrys", u"H"),
    ("MilliHenrys", u"mH"),
    ("Farads", u"F"),
    ("MilliFarads", u"mF"),
    ("MicroFarads", u"\u00B5F"),
    ("PicoFarads", u"pF"),

    ("Ohms", u"\u2126"),
    ("Siemens", u"S"),

    ("Moles", u"mol"),
    ("Becquerels", u"Bq"),
    ("Parts per Million", u"ppm"),

    ("Decibels", u"dB"),
    ("Tenths of Decibels", u"1/10 dB"),

    ("Grays", u"Gy"),
    ("Sieverts", u"Sv"),

    ("MilliWatts", u"mW"),

    ("DBm", u"dBm"),

    ("Bytes per Second", u"B/s"),
    ("KiloBytes per Second", u"kB/s"),
    ("MegaBytes per Second", u"MB/s"),
    ("GigaBytes per Second", u"GB/s"),

    ("BTU per Hour", u"BTU/h"),

    ("PCI clock cycles", u"PCI clock cycles"),

    ("250 NanoSeconds", u"250 ns"),

    ("Us", u"U"),

    ("Amps at 120 Volts", u"A (@120V)"),

    ("Clock Ticks", u"clock ticks"),

    ("Packets", u"packets"),
    ("per Thousand Packets", u"/1000 packets"),

    # Additional values found in the DMTF CIM Schema (as of its version 2.49)
    # that are not defined as valid values in Annex C.2 of DSP0004 version 2.8
    ("-dBm", u"-dBm"),
    ("Blocks", u"blocks"),
    ("Percentage", u"%"),
    ("Proportion", u"proportion"),
    ("Tenths of Revolutions per Minute", u"1/10 RPM"),
]


@pytest.mark.parametrize(
    "units, exp_result",
    TESTCASES_SIUNIT_UNITS)
def test_siunit_units(units, exp_result):
    """
    Test function for siunit(), that test only good cases with Units values.
    """

    # The code to be tested
    result = siunit(units=units)

    assert result == exp_result



"""
All tests for pywbem.

There are different kinds of tests, organized by subdirectory:

* unittest: Module-level tests.
* functiontest: Mock the client at the HTTP level and execute operations
  against the WBEMConnection class.
* end2endtest: Run against a real WBEM server.
* manualtest: Some scripts that are typically run against real WBEM servers.

In addition, the following subdirectories exist that do not contain tests:

* dtd: The DMTF CIM-XML DTD, used by some CIM-XML related unit tests.
* schema: DMTF CIM Schema, used by some MOF compiler related unit tests.
* server_definitions: Definitions of real WBEM servers to test against,
  used by end2end tests.
* profiles: Definition of attributes of DMTF management profiles, used by
  some testcases in end2end tests.
"""

from __future__ import absolute_import, print_function
import os
from .utils import import_installed
pywbem = import_installed('pywbem')  # pylint: disable=invalid-name

print("Testing pywbem version {0} from {1}".
      format(pywbem.__version__, os.path.dirname(pywbem.__file__)))
